/* 
	Assembler routines for Abcbus project.
	This file is for PIC32 assembler preprocessor and must be named .S
	http://ww1.microchip.com/downloads/en/DeviceDoc/51833A.pdf
	Changed vector spacing from 1 to 2 (OK since there is no vector 54 or higher) 
	Spacing is n*32 bytes, i.e. n*8 instructions, i.e. 16 instructions. 
	July 2012 /A.J

	PortA <==> 8-bit data bus
	PortD <=== 16-bit address bus
	INT1 <---- XMEMFL#
	INT2 <---- XMEMW#

	Using C-variables for Z80 RAM emulation:
	abcMem[65536];
	WEmap[256];
	set WEmap[x] = 1 to enable write to abcMem[256*n + 0...255]
	
	All interrupts must use IPL7 (shadow)

	TMR3Handler must come first to initialize shadow registers!
	
	(Tab space 6 looks good - right click here and select properties)
*/

	#include <p32xxxx.h>
	.list

# --------------------------------------------

	.global Int1Handler
	.global Int2Handler
	.global TMR3Handler

	.global Int3Handler
	.global Int4Handler
	.global CNHandler

	#define IOBUFSIZE	512
	.extern ioBuf, 4*IOBUFSIZE
	.extern ioHead, 4

	.extern abcMem, 65536
	.extern WEmap, 256
	.extern pCard, 64*4
	.extern pCardsInp7, 3*8
	.extern ioStuff, 2*4

# --------------------------------------------
# all port registers are in the 0xbf880000 region. 
# The following list is from comments in p32mx795f512L.h

	#define IO_BASEADDR	0xbf88
	#define trisA		0x6000
	#define portA		0x6010
	#define latA		0x6020
	#define odcA		0x6030

	#define portB		0x6050
	#define portD		0x60D0
	#define portE		0x6110
	#define latE_set		0x6128
	#define portG		0x6190

	#define ifs0_clr		0x1034
	#define ifs1_clr		0x1044
		
	#define adr_Port		portD
	#define data_Port		portA
	#define data_Lat		latA
	#define data_Odc		odcA
	#define cn6_Port		portB
	#define CN6_PINVAL	(1<<4)
	#define led_Off		latE_set
	#define LEDREDVAL		(1<<1)

	.equ	s7_pAbcMem,		s7
	.equ	s6_pWEmap, 		s6
	.equ	s5_pInp1, 		s5
	.equ	s4_ioBufData, 	s4
	.equ	s3_pIOHead, 	s3
	.equ	s2_pIOBuf, 		s2
	.equ	s1_ffff,		s1
	.equ	s0_iogp, 		s0

	.equ	t9_pSelCard, t9
	.equ	t8_pCardsInp7, t8
	.equ	t7_cardNum, t7
	.equ	t6_clrIFS0, t6
	.equ	t5_clrIFS1, t5
	.equ	t4_pIOStuff, t4

#define	CLR_IFS0BITS	sw t6_clrIFS0, ifs0_clr(s0)
#define	CLR_IFS1BITS	sw t5_clrIFS1, ifs1_clr(s0)

# unused registers are v0,v1, a0..a3, t0..t3, at
# possibly sp

# --------------- TMR3 isr ------------------------------
# Preload shadow registers with pointers and data
# This interrupt must come first and must have IPL7 (shadow)


.section .vector_12,code
	.set noreorder
	.ent TMR3Handler
TMR3Handler:
	j	tmr3isr
	nop
	.end TMR3Handler

.section .text, code
	.set noreorder
	.ent tmr3isr
tmr3isr:
	la	s7_pAbcMem, abcMem
	la	s6_pWEmap, WEmap
	li	s5_pInp1, 0
	lui	s4_ioBufData, 0xffff
	la	s3_pIOHead, ioHead
	la	s2_pIOBuf, ioBuf
	li	s1_ffff, 0xffff
	lui	s0_iogp, IO_BASEADDR

	li	t9_pSelCard, 0
	la	t8_pCardsInp7, cardsInp7	
	li	t7_cardNum, 0
	li	t6_clrIFS0, ((1<<7)|(1<<11)|(1<<15)|(1<<19))
	li	t5_clrIFS1, 1
	la	t4_pIOStuff, ioStuff

	la	a1, IEC1
	la	a0, IEC0
	lw	v0, cn6_Port(s0)
1:
	andi	v0, v0, CN6_PINVAL		# check CN6 pin
	beq	v0, zero, 1b
	lw	v0, cn6_Port(s0)

	CLR_IFS1BITS
	CLR_IFS0BITS
	sw 	t6_clrIFS0, 0(a0)		# disables T3 and enables other isr
	sw 	t5_clrIFS1, 0(a1)

	eret
	.end tmr3isr





# --------------------- MEM READ ----------------------------
.section .vector_7,code
	.set noreorder
	.ent Int1Handler
Int1Handler:
	lhu	a0, adr_Port(s0)			# read Address
	addu	a0, a0, s7_pAbcMem		# add abcMem physical address
	lbu	t1, 0(a0)				# load byte from ram
	sb	t1, data_Lat(s0)			# write low byte to dataport	
	sb	zero, data_Odc(s0)			# ODC=0
#	ori	v0, zero, 0xffff			# v1=0xffff
	lhu	t0, portE(s0)
1:
	andi	t0, t0, 0x100			# check INT1 pin
	beq	t0, zero, 1b			# jump to nearest label 1 backward
	 lhu	t0, portE(s0)
	sb	s1, data_Lat(s0)			# write 0xff to dataport
	sb	s1, data_Odc(s0)			# ODC=0xff (use open drain)
	CLR_IFS0BITS
#	sw	a0, 0(s7_pAbcMem)			# (this line is for debug)
#	ori	t0, zero, LEDREDVAL
# 	sb	t0, latEset(s0)			# red LED off	
	eret
	.end Int1Handler


# -------------------- MEM WRITE ----------------------------
.section .vector_11,code
	.set noreorder
	.ent Int2Handler 
Int2Handler:
	lhu	a0, adr_Port(s0)			# a0 = address from bus
	lbu	t1, data_Port(s0)		# t1 = data
	srl	t0, a0, 8				# t0 = address>>8
	addu	a1, t0, s6_pWEmap			# a1 = address to WEmap[PORT>>8]
	lbu	t0, 0(a1)				# t0 = WEmap data	
	beq	t0, zero, 1f			# if WEmap == 0 then goto 1 forward	
	 CLR_IFS0BITS
	addu	a1, a0, s7_pAbcMem		# a1 = address to write
	sb	t1, 0(a1)				# store byte
#	sw	a0, 0(s7_pAbcMem)			# (this line is for debug)
1:	eret
	.end Int2Handler




# ---------------------- IO strobes -------------------------
/*
INT3 <---- C1# & C2# & C4# & OUT# (ABC80 write strobes)
	RG14 <-- C1#
	RG12 <-- C2#
	RG1  <-- OUT#

INT4 <---- CS# & C3# & INP7# (ABC80 write strobes)
	RG0  <-- CS#
	RG13 <-- C3#
	RG15 <-- INP7#

CN6  <---- INP0# & INP1# (ABC80 read strobes)
	RE4 <-- INP0#
*/

	#define	OUT0_STROBEVAL	(0x0000)
	#define	CS_STROBEVAL	(0x0100)
	#define	C1_STROBEVAL	(0x0200)
	#define	C2_STROBEVAL	(0x0300)
	#define	C3_STROBEVAL	(0x0400)
	#define	C4_STROBEVAL	(0x0500)
	#define	INP7_STROBEVAL	(0x0700)
	#define	INP0_STROBEVAL	(0x0800)
	#define	INP1_STROBEVAL	(0x0900)


# --------------------- IO-CARD struct byte offsets ---------
# pointed to by pCard + selectedCard<<2
	#define	o_pInp0Head	 0
	#define	o_pInp0end 	 4
	#define	o_pInp1	8

	#define	o_inp0stat	12
	#define	o_out0stat	14
	#define	o_c1stat	16
	#define	o_c2stat	18
	#define	o_c3stat	20
	#define	o_c4stat	22

# ioStuff[0] is set to 0xff when inp7 comes, 
# ioStuff[1] is current card number
# ioStuff[2] is set to 0xff when inp1 is not written to ioBuf for implemented card
	#define	o_inp7flag	0
	#define	o_cardNum	1
	#define	o_inp1flag	2

# ---------------------- WriteIOBuf -----------------------------------
# 0x00cc0sdd, where cc=cardNum<<2, s=strobe, dd=data
# ---------------------- exit for C1,C2,C4,OUT0, C3,CS,INP7 -----------
.section .vector_13,code
	.set noreorder
	.ent WriteIOBuf 
writeIOBuf:					# t0 must hold STROBEVAL and v1 data
	ins	t0, t7_cardNum,16,8
writeIOBuf_1:				# ... and cardNum
	or	s4_ioBufData, t0, v1
writeIOBuf_2:				# s4 = data to write
	lw	a1, 0(s3_pIOHead)		# read ioHead
writeIOBuf_3:				# a1 = ioHead
	sll	a0, a1, 2			# 4 byte per data
	addu	a0, a0, s2_pIOBuf
	sw	s4, 0(a0)			# save to ioBuf
	addiu	a1, a1, 1
	andi	a1, (IOBUFSIZE-1)
	sw	a1, 0(s3_pIOHead)		# write ioHead
	CLR_IFS0BITS
	eret
	.end WriteIOBuf 


# --------------------------------------------------------------
# --------------------- C1, C2, C4, OUT0 -----------------------
/*
INT3 <---- C1# & C2# & C4# & OUT# (ABC80 write strobes)
	RG14 <-- C1# (OUT2,x)
	RG12 <-- C2# (OUT3,x)
	RG1  <-- OUT# (OUT0,x)
*/
	#define	C1_PINVAL	(1<<14)
	#define	C2_PINVAL	(1<<12)
	#define	OUT0_PINVAL (1<<1)
	
.section .vector_15,code
	.set noreorder
	.ent Int3Handler
Int3Handler:
	lhu	v0, portG(s0)			# v0 = strobes
	lbu	v1, data_Port(s0)			# v1 = data
	andi	t0, v0, OUT0_PINVAL
	beq	t0, zero, out0isr
 	 andi	t0, v0, C1_PINVAL
	beq	t0, zero, c1isr
	 andi	t0, v0, C2_PINVAL
	beq	t0, zero, c2isr
	 sll	t0, t7_cardNum, 16
#c4
	beq	t9_pSelCard, zero, writeIOBuf_1
	 ori	t0, t0, C4_STROBEVAL
	b	c4isr
	 or	s4_ioBufData, t0, v1		# s4 = cardnum, strobe, data
	.end Int3Handler


.section .vector_14, code
	.set noreorder
	.ent c4isr
c4isr:						# --------- C4 (OUT5,x) ---------
	lbu	v0, 0(s5_pInp1)
	lhu	t1, o_c4stat(t9_pSelCard)	# c4status
	and	v0, v0, t1
	srl	t1, t1, 8
	or	t1, t1, v0
	sb	t1, 0(s5_pInp1)
	b	writeIOBuf_3
	 lw	a1, 0(s3_pIOHead)			# read ioHead
	.end c4isr


.section .vector_16, code
	.set noreorder
	.ent out0isr
out0isr:					# --------- OUT0 -----------
	beq	t9_pSelCard, zero, writeIOBuf
	 ori	t0, zero, OUT0_STROBEVAL
	lbu	v0, 0(s5_pInp1)
	lhu	t1, o_out0stat(t9_pSelCard)	# out0status
	and	v0, v0, t1
	srl	t1, t1, 8
	or	t1, t1, v0
	sb	t1, 0(s5_pInp1)
	b	writeIOBuf_1
	 ins	t0, t7_cardNum,16,8
	.end out0isr


.section .vector_17, code
	.set noreorder
	.ent c1isr
c1isr:						# --------- C1 (OUT2,x) -------
	beq	t9_pSelCard, zero, writeIOBuf
	 ori	t0, zero, C1_STROBEVAL
	lbu	v0, 0(s5_pInp1)
	lhu	t1, o_c1stat(t9_pSelCard)	# c1status
	and	v0, v0, t1
	srl	t1, t1, 8
	or	t1, t1, v0
	sb	t1, 0(s5_pInp1)
	b	writeIOBuf_1
	 ins	t0, t7_cardNum,16,8
	.end c1isr


.section .vector_18, code
	.set noreorder
	.ent c2isr
c2isr:						# --------- C2 (OUT3,x) --------
	beq	t9_pSelCard, zero, writeIOBuf_1
	 ori	t0, t0, C2_STROBEVAL
	lbu	v0, 0(s5_pInp1)
	lhu	t1, o_c2stat(t9_pSelCard)	# c2status
	and	v0, v0, t1
	srl	t1, t1, 8
	or	t1, t1, v0
	sb	t1, 0(s5_pInp1)
	b	writeIOBuf_2
	 or	s4_ioBufData, t0, v1
	.end c2isr


#----------------------------------------------------------
# --------------------- CS, C3, INP7 ----------------------
/*
INT4 <---- CS# & C3# & INP7# (ABC80 write strobes)
	RG0  <-- CS# (OUT1,x)
	RG13 <-- C3# (OUT4,x)
	RG15 <-- INP7#
*/
	#define	CS_PINVAL	(1<<0)
	#define	C3_PINVAL	(1<<13)
	#define	INP7_PINVAL	(1<<15)

.section .vector_19,code
	.set noreorder
	.ent Int4Handler 
Int4Handler:
	lhu	v0, portG(s0)			# v0 = strobes
	lbu	v1, data_Port(s0)			# v1 = data
#	lbu	v0, 0(s5_pInp1)			# (for testing exception handler)
	andi	t0, v0, INP7_PINVAL		# test if strobe is low
	beq	t0, zero, inp7isr
	 lw	a0, 0(t8_pCardsInp7)
	andi	t0, v0, CS_PINVAL
	bne	t0, zero, c3isr
	 ori	t0, zero, C3_STROBEVAL
	b	csisr
	 andi	t7_cardNum, v1, 63	 
	.end Int4Handler	


.section .vector_20, code
	.set noreorder
	.ent csisr
csisr:					      # --------- CS (OUT1,x) ------
	or	s4_ioBufData, zero, s1_ffff	# forget previous ioBuf data
	sll	t0, t7_cardNum, 2			# t0 = (cardNum & 63) << 2
	la	a0, pCard
	addu	a0, a0, t0
	lw	t9_pSelCard, 0(a0)		# pSelCard = *(pCard + t0)
	sb	t7_cardNum, o_cardNum(t4_pIOStuff)
	beq	t9_pSelCard, zero, 1f
	 CLR_IFS0BITS
	lw	s5_pInp1, o_pInp1(t9_pSelCard)
1:	eret
	.end csisr


.section .vector_21, code
	.set noreorder
	.ent c3isr
c3isr:						# --------- C3 (OUT4,x) ----
	beq	t9_pSelCard, zero, writeIOBuf_1
	 ins	t0, t7_cardNum,16,8
	lbu	v0, 0(s5_pInp1)
	lhu	t1, o_c3stat(t9_pSelCard)	# c3status
	and	v0, v0, t1
	srl	t1, t1, 8
	or	t1, t1, v0
	sb	t1, 0(s5_pInp1)
	b	writeIOBuf_2
	 or	s4_ioBufData, t0, v1
	.end c3isr



.section .vector_22, code
	.set noreorder
	.ent inp7isr
inp7isr:						# --------- INP7 ----------
	ori	t9_pSelCard, zero, 0
	ori	t7_cardNum, zero, 0
	sh	s1_ffff, o_inp7flag(t4_pIOStuff)	#sets inp7flag and cardNum
#	lw	a0, 0(t8_pCardsInp7)		# Done at jump to inp7isr
	lbu	t1, 4(t8_pCardsInp7)
	sb	t1, 0(a0)

	lw	a0, 8(t8_pCardsInp7)
	lbu	t1, 12(t8_pCardsInp7)
	sb	t1, 0(a0)				# set inp1

	lw	a0, 16(t8_pCardsInp7)
 	lbu	t1, 20(t8_pCardsInp7)
	sb	t1, 0(a0)				# set inp1

	ori	t0, zero, INP7_STROBEVAL
	bne	t0, s4_ioBufData, writeIOBuf_2	# skip repeated inp7
	 or	s4_ioBufData, zero, t0
	CLR_IFS0BITS
	eret
	.end inp7isr


# ----------------------------------------------------------
# --------------------- INP0, INP1 -------------------------
/*
CN6 (RB4)  <---- INP0# & INP1# (ABC80 read strobes)
	RE4 <-- INP0#
*/
	#define	INP0_PINVAL (1<<4)
	#define	CN6_PINVAL	(1<<4)

.section .vector_26,code
	.set noreorder
	.ent CNHandler 
CNHandler:
	lhu	v0, portE(s0)				# v0 = inp0 strobe
	andi	t0, v0, INP0_PINVAL
	bne	t0, zero, Inp1isr				# branch to .text might not work
	 sll	t0, t7_cardNum, 16			# t0 = cardNum
	j	Inp0isr
	 ori	t0, t0, INP0_STROBEVAL			# t0= card + strobe
	.end CNHandler
	
.section .text,code
	.set noreorder
	.ent Inp0isr
Inp0isr:						# ---------- INP0 ------------
	beq	t9_pSelCard, zero, Inp0_extern
	 nop
	lw	a0, o_pInp0Head(t9_pSelCard)	# a0 is current inp0 pointer
	lbu	v1, 0(a0)
 	sb	v1, data_Lat(s0)			# write low byte to dataport	
	sb	zero, data_Odc(s0)		# ODC=0
	lw	a1, o_pInp0end(t9_pSelCard)	# a1=end of inp0 buffer
	beq	a0, a1, Inp0_writeInp1		# inp1 update at end of buffer
 	 or	s4_ioBufData, t0, v1		# ready to save to ioBuf
	addiu	a0, a0, 1				# next byte
	j	Inp0_writeIObuf
	 sw	a0, o_pInp0Head(t9_pSelCard)	# save updated pointer
Inp0_writeInp1:
	lhu	t1, o_inp0stat(t9_pSelCard)	# inp0status
	lbu	v0, 0(s5_pInp1)			# current inp1
	and	v0, v0, t1
	srl	t1, t1, 8
	or	t1, t1, v0
	sb	t1, 0(s5_pInp1)			# update inp1
Inp0_writeIObuf:
	lw	a1, 0(s3_pIOHead)			# read ioHead
	sll	a0, a1, 2
	addu	a0, a0, s2_pIOBuf
	sw	s4_ioBufData, 0(a0)		# save to ioBuf
	lw	v0, cn6_Port(s0)
1:
	andi	v0, v0, CN6_PINVAL		# check CN6 pin
	beq	v0, zero, 1b
	 lw	v0, cn6_Port(s0)

	sb	s1_ffff, data_Lat(s0)		# write 0xff to dataport
	sb	s1_ffff, data_Odc(s0)		# ODC=0xff (activate open drain)
	addiu	a1, a1, 1
	andi	a1, (IOBUFSIZE-1)
	sw	a1, 0(s3_pIOHead)			# write ioHead
	CLR_IFS1BITS
	eret

Inp0_extern:					#t0=card+strobe
	lw	a1, 0(s3_pIOHead)			# read ioHead
	sll	a0, a1, 2
	addu	a0, a0, s2_pIOBuf
	lw	v0, cn6_Port(s0)
1:
	lbu	v1, data_Port(s0)			# v1 = data
	andi	v0, v0, CN6_PINVAL		# check CN6 pin
	beq	v0, zero, 1b
	 lw	v0, cn6_Port(s0)

 	or	s4_ioBufData, t0, v1
	sw	s4_ioBufData, 0(a0)		# save to ioBuf
	addiu	a1, a1, 1
	andi	a1, (IOBUFSIZE-1)
	sw	a1, 0(s3_pIOHead)			# write ioHead
	CLR_IFS1BITS
	eret
	.end Inp0isr		#----------------------------------------


.section .vector_30,code
	.set noreorder
	.ent Inp1isr
Inp1isr:						# ---------- INP1 ------------
	beq	t9_pSelCard, zero, 1f
	 ori	t0, t0, INP1_STROBEVAL		# t0 = card + strobe

	lbu	v1, 0(s5_pInp1)
 	sb	v1, data_Lat(s0)			# write low byte to dataport	
	sb	zero, data_Odc(s0)		# ODC=0
	j	Inp1_cont
 	 ins	t0, v1, 0, 8			# t0 = data to save to ioBuf
	
1:	j	Inp1_extern
	 nop
	.end Inp1isr


.section .text, code
	.set noreorder
	.ent Inp1_cont
Inp1_cont:						# t0 = data for ioBuf
	beq	t0, s4_ioBufData, 2f		# don't update ioBuf for repeated inp1
	lw	a1, 0(s3_pIOHead)			# read ioHead
	sll	a0, a1, 2
	addu	a0, a0, s2_pIOBuf			# ready to save
	sw	t0, 0(a0)				# save to ioBuf
	or	s4_ioBufData, zero, t0 		# remember latest ioBuf data
 	addiu	a1, a1, 1
	andi	a1, (IOBUFSIZE-1)
	sw	a1, 0(s3_pIOHead)			# save ioHead
	lw	v0, cn6_Port(s0)
1:
	andi	v0, v0, CN6_PINVAL		# check CN6 pin
	beq	v0, zero, 1b
	 lw	v0, cn6_Port(s0)
	sb	s1_ffff, data_Lat(s0)		# write 0xff to dataport
	sb	s1_ffff, data_Odc(s0)		# ODC=0xff (open drain)
	CLR_IFS1BITS
	eret
2:							# repeated inp1 for implemented card
	sb	s1_ffff, o_inp1flag(t4_pIOStuff)
	lw	v0, cn6_Port(s0)
1:
	andi	v0, v0, CN6_PINVAL		# check CN6 pin
	beq	v0, zero, 1b
	 lw	v0, cn6_Port(s0)
	sb	s1_ffff, data_Lat(s0)		# write 0xff to dataport
	sb	s1_ffff, data_Odc(s0)		# ODC=0xff (open drain)
	CLR_IFS1BITS
	eret


Inp1_extern:					#t0=card+strobe
	lw	a1, 0(s3_pIOHead)			# read ioHead
	sll	a0, a1, 2
	addu	a0, a0, s2_pIOBuf
	lw	v0, cn6_Port(s0)
1:
	lbu	v1, data_Port(s0)			# v1 = data
	andi	v0, v0, CN6_PINVAL		# check CN6 pin
	beq	v0, zero, 1b
	 lw	v0, cn6_Port(s0)

	ins	t0, v1, 0, 8
	beq	t0, s4_ioBufData, 2f		# skip repeated inp1	
	 CLR_IFS1BITS
	or	s4_ioBufData, zero, t0
	sw	s4_ioBufData, 0(a0)		# save to ioBuf
	addiu	a1, a1, 1
	andi	a1, (IOBUFSIZE-1)
	sw	a1, 0(s3_pIOHead)			# save ioHead
2:
	eret

	.end Inp1_cont

# end
