/*
Hardware init for ABCbus (UBW32 PIC32MX795F512L).

PortA <==> 8-bit data bus
PortD <=== 16-bit address bus
INT1 <---- XMEMFL#
INT2 <---- XMEMW#
INT3 <---- C1# & C2# & C4# & OUT# (ABC80 write strobes)
	RG14 <-- C1#
	RG12 <-- C2#
	RG1  <-- OUT#

INT4 <---- CS# & C3# & INP7# (ABC80 write strobes)
	RG0  <-- CS#
	RG13 <-- C3#
	RG15 <-- INP7#

CN6  <---- INP0# & INP1# (ABC80 read strobes)
	RE4 <-- INP0#

UART2   <--> WiFi module
UART1   <--> TTL output (connect to FTDI USB cable)
USBmini <--> PC Virtual Com Port
SPI (by software) <---> pic16f1455 (not mounted)
*/

#include "common.h"

//PORTA: INT4,INT3,0,0, 0,wifiRST,wifiGPIO4,0, d7...d0
//Default tris = 0xc6ff
//inputs tris = 0b1100 0010 0000 0000 = 0xc200
//opendrain = 0b0000 0000 1111 1111 = 0x00ff
//#define initTrisA	(0xc2ff)
#define initTrisA	(0xc200)
#define initLatA	(0xffff)
#define initodcA	(0x00ff)

//PORTB: RB15,RB14,RB13,RB12, RB11,RB10,RB9,RB8, PGD2,PGC2,softSS,CN6#, RB3,RB2,memEn,RB0
//Connect memEn to RB0 to disable memory emulation (int1, int2)
//Default tris = 0xffff
//inputs tris = 0b0000 0000 0001 0010 = 0x0012
//opendrain = 0b0000 0000 0000 0000 = 0x0000
#define initTrisB	(0x0012)
#define initLatB	(0x0000)
#define initodcB	(0x0000)

//PORTC: osc2,cts1,rts1,osc1, 0000, 000,SDcardDetect, softSDI,softSDO,softSCK,0
//Default tris = 0xf01e (note that cts1 and rts1 were swapped in schematic)
//inputs tris = 0b1101 0000 0001 1000 = 0xb018
//opendrain = 0b0000 0000 0000 0000 = 0x0000
#define initTrisC	(0xd018)
#define initLatC	(0x2000)
#define initodcC	(0x0000)

//PORTD: input from ABC80 address bus
//Default tris = 0xffff
//inputs tris = 0b1111 1111 1111 1111 = 0xffff
//opendrain = 0b0000 0000 0000 0000 = 0x0000
#define initTrisD	(0xffff)
#define initLatD	(0xffff)
#define initodcD	(0x0000)

//PORTE: 0000, 00,INT2#,INT1#, BTNprg,BTNuser,abcRESIN,inp0, LEDgrn,LEDwht,LEDred,LEDyel
//Default tris = 0x03ff
//inputs tris = 0b0000 0011 1111 0000 = 0x03f0	(set ABCresin as output=0 to reset ABC80) 
//opendrain = 0b0000 0000 0010 0000 = 0x0020
#define initTrisE	(0x03f0)
#define initLatE	(0x0320)
#define initodcE	(0x0020)

//PORTF: 00,RTS2,CTS2, 000,Tx1, 00,Tx2,Rx2, USBID,Rx1,abcNMI,abcRDY
//Default tris = 0x313f
//inputs tris = 0b0001 0000 0001 1111 = 0x101f	(set NMI and RDY as output if implemented) 
//opendrain = 0b0000 0000 0000 0011 = 0x0003
#define initTrisF	(0x101f)
#define initLatF	(0x2123)
#define initodcF	(0x0003)

//PORTG: INP7,C1,C3,C2, 0,0,SS,SDO, SDI,SCK,0,0, D+,D-,OUT,CS
//Default tris = 0xf3cf
//inputs tris = 0b1111 0000 1000 1111 = 0xf08f 
//opendrain = 0b0000 0000 0000 0011 = 0x0000
#define initTrisG	(0xf08f)
#define initLatG	(0xf3cf)
#define initodcG	(0x0000)

void hw_init(void)
{
	IEC0=0; IEC1=0; IEC2=0;
    DDPCONbits.JTAGEN = 0;
    AD1PCFG=0xffff;		//No analog pins
	LATA = initLatA;	ODCA = initodcA;	TRISA = initTrisA;
	LATB = initLatB;	ODCB = initodcB;	TRISB = initTrisB;
	LATC = initLatC;	ODCC = initodcC;	TRISC = initTrisC;
	LATD = initLatD;	ODCD = initodcD;	TRISD = initTrisD;
	LATE = initLatE;	ODCE = initodcE;	TRISE = initTrisE;
	LATF = initLatF;	ODCF = initodcF;	TRISF = initTrisF;
	LATG = initLatG;	ODCG = initodcG;	TRISG = initTrisG;
    CNPUE = (1<<17) | (1<<3);	//pullup on RF4 (U2RX) and CN3 (RB1).
	CNEN = (1<<6);	//enable CN6 (RB4)
	CNCONbits.ON = 1;

	SYSTEMConfigPerformance(80000000UL); //optimize waitstates, cache and stuff

	//All interrupts need highest prio to use shadow (cannot use nested interrupts)
  	IPC3bits.T3IP = 7;		//TMR3 isr is used for initialization of shadow registers
   	IPC3bits.INT3IP = 7;	//ABC80 io
   	IPC4bits.INT4IP = 7;	//ABC80 io
   	IPC6bits.CNIP = 7;		//ABC80 io
	if(PORTBbits.RB1)
	{
	   	IPC1bits.INT1IP = 7;	//ABC80 MEMFL_n
   		IPC2bits.INT2IP = 7;	//ABC80 MEMWR_n
	}
	else
	{
	   	IPC1bits.INT1IP = 0;	//ABC80 MEMFL_n disabled
   		IPC2bits.INT2IP = 0;	//ABC80 MEMWR_n disabled
	}

	//T4CON = 0x0060;	//16bit timer, 0.8 s tick.
	//PR4 = 0xffff;
	//T4CONSET = 0x8000; //start
	T4CON=0;
	T5CON=0;
	T4CONSET = 0x0068;	//32bit, 0.8 us tick
	TMR4=0;
	PR4=0xffffffff;
	T4CONSET=0x8000;	//start
  	
   	INTEnableSystemMultiVectoredInt();

	//uarts initialized in uarts.c and usb.c

	LEDRED_OFF;
}

//end
