/*
This is disk controller card 36 emulation
Expected ABC80 DOS sequence:
1. OUT 1, 36 (card select).
2. Wait for status (INP1) bit7=1, or else send command C3 (max 3 times).
3. Send C1-command OUT 2, 36
4. Wait for status = 0b******01 (try max 65536 times).
5. Send 4 byte command OUT 0, x.
6. Wait for status = 0b0****0*1 (forever). (if bit7=1 and inp0=0 then abort) 
7. Read or Write 256 byte data.
8. Wait for status bit7=1 (forever).
9  Read Aux_status = 0 (INP0).
*/

#include "common.h"
#include "abcio.h"
#include "uarts.h"
#define FATFS_NO_DEF_TYPES
#include "SDMMC.h"
#include "fat_filelib.h"

uint8 filebuf[256];
uint8 cmd[5];
uint8 cmdidx, dataidx;
FL_FILE *abcdiskfile;
#define ABCDISK_FILEPATH	"/abcdisk.hd0"
Bool disk36open = False;

s_card disk36 = 
{
	(uint8*) &filebuf[255],		//inp0 pointer
	(uint8*) &filebuf[255],		//end of inp0 buffer
	(uint8*) &disk36.myReg[1],	//point to status
	0x00ff,				//inp0stat=0x00ff --> inp0 does not change inp1 status
	0x7e00,				//out0stat=0x7e00 --> out0,x gives inp1 = 0x7e
	0x0100,				//out2 (C1) gives inp1=1
	0x00ff,				//out3 (C2) not implemented
	0x8100,				//out4 (C3) gives inp1 = 0x81
	0x00ff,				//out5 (C4) not implemented
	diskFunc,
	{0,0,0,0,0,0,0,0}	//myReg[8]
};

	#define card	disk36

	#define inp0reg 	filebuf[255]
	#define inp1reg 	card.myReg[1]
	#define out0reg 	card.myReg[2]
	#define out2reg 	card.myReg[3]
	#define out3reg 	card.myReg[4]
	#define out4reg 	card.myReg[5]
	#define out5reg 	card.myReg[6]
	#define inp7reg 	card.myReg[7]
	#define inp0stat	card.inp0stat
	#define out0stat	card.out0stat
	#define out2stat	card.out2stat
	#define pInp0		card.pInp0

//-----------------------------------------------------------------------
void disk36_close(void)
{
	disk36open = False;
	inp0stat = 0x00ff;
	out0stat = 0x00ff;
	inp0reg = 0x00;	
	inp1reg = 0x81;
	pInp0 = &inp0reg;
	//fl_fclose(abcdiskfile);
	abcdiskfile = NULL;
}

//-----------------------------------------------------------------------
void disk36_open(void)
{
	inp0stat = 0x00ff;
	abcdiskfile=fl_fopen(ABCDISK_FILEPATH, "r+");
	if(abcdiskfile != NULL)
	{
		inp0reg = 0;	
		inp1reg = 0x80;	
		out0stat = 0x7e00;		//sets inp1reg=0x7e when out0,x comes
		disk36open = True;
	}
	else
	{
		out0stat = 0x00ff;		//out 0,x does not change inp1reg
		inp0reg = 0x00;	
		inp1reg = 0x81;
		disk36open = True;
	}
	pInp0 = &inp0reg;
	cmdidx=0;
	dataidx=0;
	cmd[0] = 0;
}

//-----------------------------------------------------------------------
void readyForCommand(void)
{
	inp0stat = 0x00ff;
	inp1reg = 0x81;		//Done
	inp0reg = 0;
	pInp0 = &inp0reg;
	cmdidx=0;
	cmd[0] = 0;
}

//-----------------------------------------------------------------------
void diskFunc(uint8 strobe, uint8 data)
{
	uint16 i;

	if(disk36open == False) 
	{
		if(ioprint) printf("SD not ready\n");
		inp0reg = 0x00;
		inp1reg = 0x81;
		return;
	}

	switch(strobe)
	{
	case STROBE_OUT0:			// gives status = 0x7e (not ready)
		if(cmdidx==0)
		{
			inp0reg = 0;
			inp0stat = 0x00ff;
			dataidx=0;
			if(abcdiskfile==NULL)
				abcdiskfile=fl_fopen(ABCDISK_FILEPATH, "r+");
		}
		if (cmdidx<4) 
		{
			cmd[cmdidx++]=data; 
			if(cmdidx<4) 
			{
				inp1reg = 0x01;	// ready for next byte
				//if(ioHead == ioTail) delay_us(5);	// just wait a little for next byte
				break;
			}

			//we have received a command
			if(ioprint) 
			{
				while(serviceTx1());	//empty UART1 Tx buffer
			}

			if(ioprint == 2)
				printf("\nCmd %d %d %d %d %d\n",cmdidx,cmd[0],cmd[1],cmd[2],cmd[3]);

			if (abcdiskfile==NULL)
			{
				inp0reg = 0x00;
				inp1reg = 0x80;
				if(ioprint) printf("abcdiskfile not found!");
				break;
			}

			if (cmd[0]==3) 
			{ 	/// a read command
				//System.out.println("Seek and read: T:"+cmd[2]+" S:"+cmd[3]+" equals offset "+Integer.toHexString((cmd[3]*256+256*256*cmd[2])));
				if(ioprint == 2)
					printf("READ drive %d T:%d S:%d\n",cmd[1],cmd[2],cmd[3]);
				//rb->lcd_putsxy(40,27*8,dtext);
			     
				/// the disk file is 4 disks * 238 tracks * 256 sectors * 256 bytes in size
				/// this is the largest possible size (according to HPA's investigations)
				/// cmd[1] holds the drive number (0-3) (plus possibly some other junk which is filtered away
				/// cmd[2] holds the track number and cmd[3] the sector number
	
				// this is for the harddisk
				i = fl_fseek(abcdiskfile,(cmd[1] & 3)*238*256*256 + cmd[2]*256*256 + cmd[3]*256, SEEK_SET);
				// this is for ABC830 only one drive at the moment!
				// abc830 has 1 head 40 tracks and 16 sectors per track
				// fseek(abcdiskfile,(cmd[3] & 15)*256+(cmd[2]*16*256),SEEK_SET);

				if(i != 0)
				{
					inp0reg = 0x10;	
					inp1reg = 0x88;
					if(ioprint) printf("Seek error! %d", i);
					break;
				}

				i = fl_fread(filebuf,256,1,abcdiskfile);
			    if(i != 256)
				{
					inp0reg = 0x08;	
					inp1reg = 0x88;
					if(ioprint) printf("Read error! %d bytes", i);
					break;
				}
			    //for (i=0;i<16;i++) printf("_ %d", *pInp0);
			    //  System.out.print(" "+Integer.toHexString(filebuf[i]&0xff));
			    pInp0 = filebuf;
				inp0stat = 0;
			    inp1reg = 1; 		// data is ready
				cmd[0] = 2;			// buffer to host remains
			}
			else if (cmd[0] == 12)
			{
				inp1reg=1;	//ready to receive data
				if(ioHead == ioTail) delay_us(5);
			}
			else
			{
				inp0reg = 0x08;	
				inp1reg = 0x88;
				if(ioprint) printf("Unknown command! %d", cmd[0]);
				cmd[0] = 0;
			}

			break;
		}

		if (cmd[0]==12) 
		{
			/// a write command
	 		filebuf[dataidx++]=data;
			if(dataidx != 0) 
			{	//less than 256 byte received
				//if(ioHead == ioTail) delay_us(5);
				break;
			}

			if (abcdiskfile==NULL) break;	
			/// the disk file is 4*238*256*256 bytes in size which yields 4 disks of 238 tracks of 256 sectors with 256 bytes
			/// this is the largest possible size (according to HPA)					
			if (ioprint == 2) 
				printf("\n\rWRITE drive:%d T:%d S:%d",cmd[1],cmd[2],cmd[3]);
		
			// this is for the harddisk
			i = fl_fseek(abcdiskfile,(cmd[1] & 3)*238*256*256 + cmd[2]*256*256 + cmd[3]*256, SEEK_SET);
			if(i != 0)
			{
				inp0reg = 0x10;	
				inp1reg = 0x80+0x10;
				if(ioprint) printf("Seek error! %d", i);
				break;
			}

			// this is for ABC830 only one drive at the moment!
			//fseek(abcdiskfile,(cmd[3] & 15)*256+(cmd[2]*16*256),SEEK_SET);
				
			i = fl_fwrite(filebuf,256,1,abcdiskfile);
			if(i != 256)
			{
				inp0reg = 0x40;
				inp1reg = 0x80+0x40;
				if(ioprint) printf("Write fault! %d", i);
				break;
			}
			fl_fflush(abcdiskfile); // flush if everything was OK!
			inp0reg = 0;
			inp1reg = 0x80; 		// Done
			cmd[0] = 0;
		}
		break;
		
	case STROBE_CS:		//OUT 1, this card
		break;

	case STROBE_OUT2:	 //C1 command: set controller in command mode and reset pointers
		//out2reg = data;
		cmdidx = 0;
		inp0reg = 0;
		if(ioprint) { putcPC('_'); putcPC('2'); }
		break;

	case STROBE_OUT3:	
		//out3reg = data;	//not used
		//readyForCommand();
		break;

	case STROBE_OUT4:	//C3 command: reset controller
		//out4reg = data;
		readyForCommand();
		if(ioprint==2) { putcPC('_'); putcPC('4'); }
		break;

	case STROBE_OUT5:	
		//out5reg = data;
		//readyForCommand();
		if(ioprint==2) { putcPC('_'); putcPC('5'); }
		break;

	case STROBE_INP7:
		readyForCommand();
		if(ioprint==2) { putcPC('_'); putcPC('7'); }
		break;

	case STROBE_INP0:
		if(cmd[0] == 2)
		{
			dataidx++;
			if((dataidx | inp1reg) == 0)
			{	//read command finished
				readyForCommand();
				if(ioprint==2) { putcPC('_'); putcPC('z'); }
			}
		}
		break;

	case STROBE_INP1:
		if(ioprint==2) { putcPC('_'); putcPC('9'); }
		break;

	case STROBE_INP1X:
		//we may have endless INP1. Assume we are Done.
		inp0stat = 0x00ff;
		inp0reg = 0;
		inp1reg = 0x80;
		pInp0 = &inp0reg;
		cmdidx=0;
		if(ioprint==2) printf("_inp1x");
		break;

	default:
		break;
	}
}

//end
