/* Program ABCbus for UBW32
** MPLAB 8.92, PIC32-GCC v2.02
** Linker script for use with bootloader. 
** Note vector_spacing=2.
*/

#include <p32xxxx.h>
#include "common.h"
#include "abcMemory.h"
#include "uarts.h"

// POSCMOD = HS, FNOSC = PRIPLL, FWDTEN = OFF
// Main clock = 8MHz /2 * 20    = 80MHz
// Peripheral clock = 80MHz /1  = 80MHz

#pragma config POSCMOD = HS, FNOSC = PRIPLL, FWDTEN = OFF
#pragma config FPLLMUL = MUL_20, FPLLIDIV = DIV_2, FPLLODIV = DIV_1
#pragma config FPBDIV = DIV_1

//USB PLL Input must be 4 MHz
#pragma config UPLLIDIV = DIV_2, UPLLEN = ON
//#pragma config FVBUSONIO = OFF

//-------------------------- variables used in isr ----------------------
extern const uint8 abcRom[32768];
volatile unsigned char __attribute__((persistent)) abcMem[65536];	//will copy from abcRom[]
unsigned char WEmap[256];	//Set WEmap[0]==0 to write protect abcMem[0..255]

//----------------- functions ---------------------------

static void abcMem_init(void);
static void isr_init(void);
static void sd_detectCheck(void);

//---------------------------------------------------------

int main(void)
{
	hw_init();				//define hardware
	abcMem_init();			//initialize emulated ABC80 memory
	abcioInit();
	isr_init();				//initialize ABC80 bus interrupts
	LEDGRN_OFF;				//SD card ready
	LEDYEL_OFF;				//SD read/write
	uart_init();
	usb_init();
	sd_detectCheck();		//init SD card
	printf("\n\r%%%%%% ABCbus Interface %s %%%%%%\n\r",ABCBUS_VER);
	LEDWHT_OFF;
		
	while(1)
	{
		usb();				//USB serial com port
		sd_detectCheck();
		rjdos();			//SD card and stuff
		pollUart1();
		abcioCheck();
		pollUart2();
	}
}

//----------------------------------------------
void sd_detectCheck()
{	//check if SD is present and handle close/init. Also ABC reset button.
	#define SD_INIT_DELAY	(100*TMR4_ms)
	#define RESET_DELAY		(1000*TMR4_ms)
	//lng delay, annars detekteras extra reset nr knappen slpps
	static uint32 reset_time = 0;
	static Bool	resetDetected = False;
	static uint32 sd_notReady_time = 0;
	static Bool sd_initDelay = False;	
	static Bool sd_ready = False;
	uint32 n=0;

	if(ABCRESIN == 0)
	{
		LEDWHT_ON;
		if(! resetDetected)
		{	//New reset detected, let's just reset disk
			if(sd_ready)
			{
				abcio_disk_close();
				rjdos_sd_close();
				LEDGRN_OFF;			//SDMMC will turn ON when successfully initialised
				sd_ready = False;
				sd_initDelay = False;
			}
			resetDetected = True;
			reset_time = TMR4;
			printf("ABC reset\r\n>");
		}
	}
	else if(resetDetected)
	{
		n = TMR4 - reset_time;
		if(n > RESET_DELAY) resetDetected = False;
		LEDWHT_OFF;
		n=0;
	}

	if(SD_DETECT_PIN == 1)
	{
		sd_notReady_time = TMR4;
		sd_initDelay = True;
		if(sd_ready)
		{
			abcio_disk_close();
			rjdos_sd_close();
			LEDGRN_OFF;			//SDMMC will turn ON when successfully initialised
			sd_ready = False;
			sd_initDelay = False;
		}
	}
	else if(sd_ready == False)
	{
		if(sd_initDelay)
		{
			n = TMR4 - sd_notReady_time;
			if(n > SD_INIT_DELAY) sd_initDelay = False;
		}
		else
		{			
			n = rjdos_sd_init();
			if(n == 0)
			{
				sd_ready = True;	//SD init success
				abcio_disk_open();
			}
			sd_notReady_time = TMR4;
			sd_initDelay = True;
		}
	}
}

//------------------------------------------------------------------------
//void _general_exception_handler(unsigned cause, unsigned status)
void _general_exception_context(void)
{
	#define EXCEP_MSG "\r!exception_addr: "
	static enum {
		EXCEP_IRQ = 0,			// interrupt
		EXCEP_AdEL = 4,			// address error exception (load or ifetch)
		EXCEP_AdES,				// address error exception (store)
		EXCEP_IBE,				// bus error (ifetch)
		EXCEP_DBE,				// bus error (load/store)
		EXCEP_Sys,				// syscall
		EXCEP_Bp,				// breakpoint
		EXCEP_RI,				// reserved instruction
		EXCEP_CpU,				// coprocessor unusable
		EXCEP_Overflow,			// arithmetic overflow
		EXCEP_Trap,				// trap (possible divide by zero)
		EXCEP_IS1 = 16,			// implementation specfic 1
		EXCEP_CEU,				// CorExtend Unuseable
		EXCEP_C2E				// coprocessor 2
	} _excep_code;
	
	//static unsigned int _excep_code;
	static unsigned int _excep_addr;
	static uint32 i,j,n;

	TRISA=0xc2ff;

	asm volatile("mfc0 %0,$13" : "=r" (_excep_code));
	asm volatile("mfc0 %0,$14" : "=r" (_excep_addr));	//doesn't work in interrupt?
//	asm volatile("mfc0 %0,$30" : "=r" (_excep_addr));	//doesn't work?
//	asm volatile("mfc0 %0,$8" : "=r" (_excep_addr));

	_excep_code = (_excep_code & 0x0000007C) >> 2;

	TRISECLR=2;		//led red pin
	LEDWHT_OFF;
	LEDGRN_OFF;
	LEDYEL_OFF;

	for(j = 0; j<30; j++)
	{
		if(EXCEP_MSG[j] == 0) break;
		U1TXREG = EXCEP_MSG[j];
		for(i=0;i<100000;i++);
	}
	j = _excep_addr;
	for(n=0; n<8; n++)
	{
		i = (j>>28) + '0';
		if(i>'9') i += ('a' - '9' - 1);
		U1TXREG = i;
		for(i=0;i<100000;i++);
		j<<=4;
	}
	U1TXREG = CR;

	while(1)
	{	//flash red led approx 10 Hz
		for (i=500000; i!=0; i--);
		LEDRED_OFF;
		if(ABCRESIN == 0) SoftReset();
		for (i=500000; i!=0; i--);
		LEDRED_ON;
	}
}


//----------------------------------------------------------------------------------
void abcMem_init(void)
{
	uint32 n;
	//volatile uint8* p;
	for(n=0; n<32768; n++) abcMem[n] = abcRom[n];
	while(n < 65536) abcMem[n++] = 0xff;
	for(n=128; n<=255; n++) WEmap[n]=1;	//RAM at >= 32768
	//abcMem[65535] = 0x05;

	//ufd Ctrl-C patch frn ufdpatch.pl
	abcMem[0x6851]=0xcd;
	abcMem[0x6852]=0x9b;
	abcMem[0x6853]=0x6f;
	abcMem[0x6f9b]=0xcd;
	abcMem[0x6f9c]=0x3e;
	abcMem[0x6f9d]=0x03;
	abcMem[0x6f9e]=0xe9;
	//ufd patch 2,4,5 frn ufdpatch.ny r redan i ufddos.bin

	//The following is for rjdos
	abcMem[SDSTAT]=99; 		// SD0 not ready
	for(n=67;n<74;n++) WEmap[n]=1;	//adr 17152...

	//write ABCbus version
	#define ABCBUS_VERSION_DOT_ADDRESS	16462
	if(abcMem[ABCBUS_VERSION_DOT_ADDRESS] == '.')
	{
		abcMem[ABCBUS_VERSION_DOT_ADDRESS-1] = ABCBUS_VER[0];
		abcMem[ABCBUS_VERSION_DOT_ADDRESS+1] = ABCBUS_VER[2];
	}
}

//-----------------------------------------
void isr_init(void)
{	//initiate abcMem first
	//init ISR - trig TMR3 isr before any other interrupt!
	while(PORTBbits.RB4 == 0);	//If this pin is stuck then isr will not return
	IFS1=0;
	IFS0=0;
	IFS0bits.T3IF=1;
	IEC0bits.T3IE=1;			//T3 isr will enable the other interrupts
	while (IEC0bits.T3IE);		//T3 isr will disable T3 interrupt

//	IEC0bits.INT1IE=0;
//	IEC0bits.INT2IE=0;

	/*
 	//debug test for isr context corruption
	uint32 n,m;
	n=m=0;
	LEDGRN_ON;
	while(1)
	{       //Let's do some math to see if context is preserved during interrupt
        n++;
        m--;
        if ((n+m) != 0)
        {       //we are dead!
                LEDGRN_OFF;
        }
	}
	*/
}

//end.
