/* 
** Memory mapped SD interface for ABC80 device SD0:.
** Basically, ABC80 writes CMD and FNO and sets STAT=128.
** PIC32 checks for STAT=128, performs the CMD and sets STAT<128 when ready.
*/
#define FATFS_NO_DEF_TYPES
#include "common.h"
#include "uarts.h"
#include "SDMMC.h"
#include "fat_filelib.h"
#include "abcMemory.h"
#include <string.h>

#include "abcio.h"
//set ioprint=1 to send all bus io to PC in abcio.c

//------------------ defines ------------------------
#define MAXFILES	8
#define SDINITATTEMPTS	2

//SD0: commands (note that bit3=0)
#define CMD_OPEN	1	//; 1:  Open file 
#define CMD_PREP	2  	//; 2:	Prepare file
#define CMD_CLOSE	3	//; 3:	Close file 
#define CMD_BLIN	4	//; 4:	Read block 
#define CMD_BLOUT	5	//; 5:	Write block 
#define CMD_SEEK	6	//; 6:	SEEK
#define CMD_INIT	99	//; 99: (re)initialize
//SD0 error codes
#define ERR_NONE	0
#define ERR_OPEN	1
#define ERR_PREP	2
#define ERR_EOF		34
#define ERR_BLIN	4
#define ERR_BLOUT	5
#define ERR_CLOSE	6
#define	ERR_FNO		7 	// wrong filenumber, >7
#define ERR_MOD		8 	// wrong file mode for operation
#define ERR_INIT	99 	// SD init error

//CMD: commands (note that bit3=1)
#define CMDC_OPEN	10 	//EQU		10
#define CMDC_CLOS	11	//EQU		11
#define CMDC_PRIN	12	//EQU		12
#define CMDC_INPU	13	//EQU	    13
//CMD errors (note that sd0dev selects ABC80 error codes)
#define ERRC_PRIN 	12	//unknown command
#define ERRC_PRLIB	48	//LIB directory error
#define ERRC_INPU	13

typedef struct {
	FL_FILE	*fid;
	int		mode;
} sdfile;

Bool testWifi = False;
Bool sd_initialised = False;

//------------------ Private functions --------------------------
int media_read(unsigned long sector, unsigned char *buffer, unsigned long sector_count);
int media_write(unsigned long sector, unsigned char *buffer, unsigned long sector_count);
void convfilename(char *outstr, char *sdfnam);
uint8 processSD(uint8 cmd);		//abc SD device
uint8 processCmdc(uint8 cmd);	//adc commandline
void checkSerialCmd(void);
uint8 serialCommand(uint32 cmd32);	//for testing commands via UART

//---------------------------------------------------------------
int media_read(unsigned long sector, unsigned char *buffer, unsigned long sector_count)
{	//from SD example.c, called by pointer reference from SD functions
    unsigned long i;

    for (i=0;i<sector_count;i++)
    {
        // ...
        // Add platform specific sector (512 bytes) read code here
        //..
		readSECTOR(sector,buffer);
        sector ++;
        buffer += 512;
    }

    return 1;
}

//-----------------------------------------------------------------
int media_write(unsigned long sector, unsigned char *buffer, unsigned long sector_count)
{	//from SD example.c, called by pointer reference from SD functions
    unsigned long i;

    for (i=0;i<sector_count;i++)
    {
        writeSECTOR(sector,buffer);
        sector ++;
        buffer += 512;
    }

    return 1;
}

//----------------------------------------------------------------
int rjdos_sd_init(void)
{	//call if sd is not yet initialised
	int n = 0;
	int c = 1;
	LEDYEL_ON;
	sd_initialised = False;
	initSD();
	n=0;
	do
	{	c = initMedia();
		//putcPC((c & 1)+'0'); putcPC(' ');
		n++;
	} while((c) && (n<SDINITATTEMPTS));
	
	if(c == 0)
	{
		fl_init();
		if (fl_attach_media(media_read, media_write) != FAT_INIT_OK)
		{
			c = 1;	//fel i biblioteket
		}
	}
	LEDYEL_OFF;
	//printf("\r\nSD init attempts=%i ",n);	
	if(c != 0) c = ERR_INIT;
	else sd_initialised = True;
	abcMem[SDSTAT] = (uint8) c;
	return c;
}

//---------------------------------------------------------------
void rjdos_sd_close(void)
{
	fl_shutdown();
	sd_initialised = False;
	abcMem[SDSTAT] = ERR_INIT;
}

//---------------------------------------------------------------
void wifi2pc(void)
{
	//for manually sending commands to WiFly via usb uart.
	uint8 c;

	if(DataRdyPC())
	{
		c = getcPC();
		//putcPC(c);
		if(c=='<')
		{
			testWifi = False;
			printf("< Wifi session ended\r\n");
		}
		else
		{
			putCharTx2(c);
			if(c==CR)
			{
				printf("\nW>");
			}
		}
	}

	while(bytesInRx2())
	{
		c = getCharRx2();
		putcPC(c);
	}
}


//----------------- Commands from ABC80 ----------------------------
void rjdos(void)
{
	uint8 cmd, status;

	if(testWifi) wifi2pc();
	else checkSerialCmd();

	if((abcMem[SDSTAT] & 128)==0) return;

	abcMem[SDSTAT] = 128; // clear errors and set busy flag
	cmd = abcMem[SDCMD];
	if(cmd & 8)
	{
		status = processCmdc(cmd);
	}
	else
	{
		if(! sd_initialised) status = ERR_INIT;
		else status = processSD(cmd);
	}
	abcMem[SDSTAT]=status; 		//clear busy flag
	if(ioprint) printf(", stat=%i\r\n>",status);
}


//-----------------------------------------------------------------
void convfilename(char *outstr, char *sdfnam)
{
	int i,j=0;
	for (i=0;i<8;i++)
	{
		if (sdfnam[i]!=' ')
		{
			outstr[j++]=sdfnam[i];
		}
	}
	outstr[j++]='.'; //add a dot in filename
	for (i=8;i<11;i++)
	{
		if (sdfnam[i]==' ') break;
		outstr[j++]=sdfnam[i];
	}
	outstr[j]='\0'; // terminate with a zero char	
}

//------------------------- SD0: Device -----------------------------------
uint8 processSD(uint8 cmd)
{
	#define MODENO		0 	// Not Open
	#define	MODEWR		1
	#define MODERO		2

	static sdfile sdfiles[MAXFILES];
	static char fullname[255],filename[14];
	static uint8 fno;
	uint8 status=0;
	uint32 i;

	fno = abcMem[SDFNO];

	if(ioprint) printf("SDCMD %i, fno %i",cmd, fno);
	if (fno>7) fno=0; // abc80 do it like this

	switch (cmd)
	{
	case CMD_INIT:
		if(sd_initialised == False) status = ERR_INIT;
		if(ioprint) printf("init SDSTAT=%i",status);
		break;
				
	case CMD_OPEN:
		convfilename(filename, (char*) &abcMem[SDFNAM]);	// convert to a real file name
		strcpy(fullname,"/"); // later to include other directories
		strcat(fullname,filename);
		sdfiles[fno].fid = fl_fopen(fullname, "r+"); //returns NULL and closes if already open
		if(sdfiles[fno].fid == NULL) sdfiles[fno].fid = fl_fopen(fullname, "r+"); //Try again
		if(sdfiles[fno].fid)
			sdfiles[fno].mode = MODEWR; // open and set mode to RW
		else status = ERR_OPEN; // error opening file
		if(ioprint) printf("\r\nOpen: %s,fno=%i,SDSTAT=%i",fullname,fno,status);
		break;

	case CMD_PREP:
		convfilename(filename, (char*) &abcMem[SDFNAM]);	// convert to a real file name
		strcpy(fullname,"/"); // later to include other directories
		strcat(fullname,filename);
		sdfiles[fno].fid = fl_fopen(fullname, "w");
		if(sdfiles[fno].fid == NULL) sdfiles[fno].fid = fl_fopen(fullname, "w"); //Try again
		if (sdfiles[fno].fid) 
			sdfiles[fno].mode = MODEWR; // open and set mode to RW
		else status = ERR_PREP; // error opening file for writing
		if(ioprint) printf("\r\nPrepare: %s,fno=%i,SDSTAT=%i",fullname,fno,status);
		break;	

	case CMD_CLOSE:
		if(sdfiles[fno].mode!=MODENO) fl_fclose(sdfiles[fno].fid);
		if(ioprint) printf("\r\nClose: fno=%i,SDSTAT=%i",fno,status);		
		break;

	case CMD_BLIN:		//if (sdfiles[fno].mode==MODERO) not needed if always can read!
		if ( fl_fread((void*) &abcMem[SDBUF], abcMem[BLKSZ], 1, (sdfiles[fno].fid) ) < 0 )
			status = ERR_BLIN; // something went wrong and we got -1 from fread
		else if (fl_feof(sdfiles[fno].fid))
			status = ERR_EOF; // if EOF during block read, mark EOF in SDSTAT	
		if(ioprint) printf("\r\nBlock in: fno=%i,SDSTAT=%i,bytenum=%lu",fno,status,sdfiles[fno].fid->bytenum);	
		break;

	case CMD_BLOUT:
		if ( fl_fwrite((void*) &abcMem[SDBUF], abcMem[BLKSZ], 1, sdfiles[fno].fid) < 0 ) 
			status=ERR_BLOUT; // something went wrong during write...
		else fl_fflush(sdfiles[fno].fid); // flush if everything was OK!
		if(ioprint) printf("\r\nBlock out: fno=%i,SDSTAT=%i bytenum=%lu",fno,status,sdfiles[fno].fid->bytenum);					
		break;

	case CMD_SEEK:
		i = (abcMem[BLK]+abcMem[BLK+1]*256);
		if (fl_fseek(sdfiles[fno].fid,i*abcMem[BLKSZ],SEEK_SET))
		{
			status=ERR_EOF;
		}
		if(ioprint) printf("\r\nSeek: fno=%i,SDSTAT=%i, Block=%lu, size=%i",fno,status,i,abcMem[BLKSZ]);	
		break;

	default:
		break;
	}

	return status;
}


//---------------------------------- CMD: device -----------------------------
uint8 processCmdc(uint8 cmd)
{
	#define CMDMAX		252 //max length of buffer
	static FL_DIR dirStat;
	static int dirFilenum = 0;
	static struct fs_dir_ent dirEntry;
	static uint32 activeCommand = 0;
	static uint32 i1=0,i2=0,i3=0;
	uint8 status=0;
	uint16 n;

	switch(cmd)
	{
	case CMDC_OPEN:				//E.g. Open "cmd:" asfile 1
		if(ioprint) printf("\r\nCMD open");
		if(activeCommand == 'L') fl_closedir(&dirStat);
		activeCommand=0;
		break;

	case CMDC_CLOS:
		if(ioprint) printf("\r\nCMD close");
		if(activeCommand == 'L') fl_closedir(&dirStat);
		activeCommand=0;
		break;

	case CMDC_PRIN:
		n = abcMem[SDBUF+1];	//only using first letter
		if(ioprint)
		{
			i1 = abcMem[SDBUF];	//length of string starting at SDBUF+1
			printf("\r\nCMD %u, %d, %c", (uint8) i1, n, (uint8) n);
			if(n>1) putcPC(abcMem[SDBUF+2]);
			if(n>2) putcPC(abcMem[SDBUF+3]);
			//printf("\r\n---1 %s %i %i",(char *)(abcMem+SDBUF+1),(uint8) activeCommand,abcMem[SDBUF+2]);
		}
		if(n == CR) break;		//this is just from PRINT #1, "LIB" (without semicolon)
		//new command detected
		if(activeCommand == 'L') fl_closedir(&dirStat);
		switch(n)
		{
			case 'l': 		// E.g. Print #1, "LIB"
			case 'L':
				activeCommand = 'L';
				if(! sd_initialised)
				{
					status = ERR_INIT;
					activeCommand = 0;
					break;
				}
				if(ioprint) printf("\r\nOpenDir 1");
			    if (!fl_opendir("/", &dirStat))
				{
					status = ERRC_PRLIB;	//error open dir 
					activeCommand = 0;
					break;
				}
			    dirFilenum=0;
				if(ioprint) printf("\r\nOpenDir OK!");
				i1=0; i2=0;
				break;
	
			case '*':		//OS
			case 'v': 	//ABCbus version
			case 'V':
				activeCommand = 'V';	
				if(ioprint) printf("\r\nOS ");
				break;
	
			case 'w': 
			case 'W':
				activeCommand = 'W';		//WiFi
				if(ioprint) printf("\r\nWIFI");
				break;
	
			default: 
				status = ERRC_PRIN;
			case 'R':		//rst command
			case 'r':
			case 0:
				activeCommand = 0;
				break;
			}	
		break;

	case CMDC_INPU:			// E.g. Input #1, A
		if(ioprint) printf("\r\nCMD input, activeCommand=%c",(uint8) activeCommand);
		abcMem[SDBUF]=0;
		switch(activeCommand)
		{
			case 'L':
				if(! sd_initialised) return ERR_INIT;
				if (fl_readdir(&dirStat, &dirEntry) != 0)
				{
					fl_closedir(&dirStat);
					//FL_UNLOCK(&_fs);
					activeCommand = 0;
					//ABC80 tar inte sista input nr eof kommer.
					//snprintf((char *)abcMem+SDBUF+1,CMDMAX,"     %lu Filer, %lu bytes",i1, i2);
					//abcMem[SDBUF]=strlen((char *)abcMem+SDBUF+1);
					status = ERR_EOF;	//man kunde lta ERR_EOF ges av activeCommand=0 (case 0)
					break;
				}
				if(dirEntry.cluster > 65536) n = 65535;
				else n=dirEntry.cluster;
				if (dirEntry.is_dir)
				{
					snprintf((char *)abcMem+SDBUF+1,CMDMAX,"%3d %13s %12s (0x%x)",++dirFilenum, dirEntry.filename, "<DIR>", n);
				}
				else
				{	//i1++;	//number of files
					i3 = dirEntry.size;
					//i2 += i3; //total size
					snprintf((char *)abcMem+SDBUF+1,CMDMAX,"%3d %13s %6lu bytes (0x%x)",++dirFilenum, dirEntry.filename, i3, n);
				}
				abcMem[SDBUF]=strlen((char *)abcMem+SDBUF+1);
				//if(ioprint) printf(" len=%d, %s ",abcMem[SDBUF], abcMem+SDBUF+1);
				break;
	
			case 'V':			//report ABCbus version?
				abcMem[SDBUF] = 3;
				abcMem[SDBUF+1] = ABCBUS_VER[0];
				abcMem[SDBUF+2] = ABCBUS_VER[1];
				abcMem[SDBUF+3] = ABCBUS_VER[2];
				abcMem[SDBUF+4] = CR;
				break;
	
			case 'W':			//input from WiFly SPI?
				status = ERR_EOF;
				break;
	
			case 0:				//nothing to report
			default:
				status = ERR_EOF;
				break;		
		}
		break;

	default:
		if(ioprint) printf("\r\nWhat???");
		activeCommand=0;
		break;	
	}
	return status;
}


//------------ Commands via UART - for testing ---------

#define SCBUFLEN 512
static unsigned char scBuf[SCBUFLEN];

void checkSerialCmd(void)
{
	static uint16 cmdidx=0;
	static uint32 cmd32=0;
	uint8 c;

	while(DataRdyPC())
	{
		c=getcPC();
		putcPC(c);
		scBuf[cmdidx]=(char)c;
		switch (c)
		{
		case 13:
		case 10: 
			scBuf[cmdidx]=0; // mark end of string instead of CR/LF
			if (cmdidx > 0) 
			{
				serialCommand(cmd32);
			}
			cmdidx=0;
			cmd32=0;
			putcPC(CR);
			putcPC(LF);
			if (testWifi) putcPC('W');
			putcPC('>');
			break;

		case 8:	//backspace
			if(cmdidx == 0) break;
			cmdidx--;
			if (cmdidx < 4) cmd32 >>= 8;
			break;

		default:
			if (cmdidx < 4)
			{
				cmd32 = (cmd32<<8) | c;
			}	
			if (cmdidx<(SCBUFLEN-1)) cmdidx++;
			break;
		}
	}
}

//-----------------------------------------------------------------
uint8 serialCommand(uint32 cmd32)
{
	#define CMD32_ls	(('l'<<8) | ('s'))
	#define CMD32_init	(('i'<<24) | ('n'<<16) | ('i'<<8) | ('t'))
	#define CMD32_buf	(('b'<<16) | ('u'<<8) | ('f'))

	#define CMD32_rm	(('r'<<8) | ('m'))
	#define CMD32_secr	(('s'<<24) | ('e'<<16) | ('c'<<8) | ('r'))
	#define CMD32_cat	(('c'<<16) | ('a'<<8) | ('t'))
	#define CMD32_cr	(('c'<<8) | ('r'))

	#define CMD32_wifi	(('w'<<24) | ('i'<<16) | ('f'<<8) | ('i'))
	#define CMD32_iost	(('i'<<24) | ('o'<<16) | ('s'<<8) | ('t'))
	#define CMD32_peek	(('p'<<24) | ('e'<<16) | ('e'<<8) | ('k'))
	#define CMD32_poke	(('p'<<24) | ('o'<<16) | ('k'<<8) | ('e'))

	#define CMD32_ledy	(('l'<<24) | ('e'<<16) | ('d'<<8) | ('y'))
	#define CMD32_ledr	(('l'<<24) | ('e'<<16) | ('d'<<8) | ('r'))
	#define CMD32_ledw	(('l'<<24) | ('e'<<16) | ('d'<<8) | ('w'))
	#define CMD32_ledg	(('l'<<24) | ('e'<<16) | ('d'<<8) | ('g'))

	#define CMD32_iopr	(('i'<<24) | ('o'<<16) | ('p'<<8) | ('r'))
	#define CMD32_boot	(('b'<<24) | ('o'<<16) | ('o'<<8) | ('t'))

	#define MAXARGS 20
	static uint32 sector = 0;
	static FL_FILE *file;
	//static struct fatfs fs;
	static char *argv[MAXARGS]; // max 20 args
	int32 i;
	int32 j, argc=0;

	//unsigned char data[100];		//CMD32_cr
	const char d2[4]={1,2,3,4};		//CMD32_cr

	argv[argc]=strtok((char*)scBuf," \t");
	while ((argc<(MAXARGS-1) && (argv[argc]!=NULL)))
		argv[++argc]=strtok(NULL," ");
	
	printf("\r\nCommand:%s - number of arguments: %ld\r\n",argv[0],argc);
	if (argc<0) return -1;

	switch(cmd32)
	{
	case CMD32_ls:
		fl_listdirectory(argv[1]);
		break;

	case CMD32_init:
		initSD();	//This will invalidate any open file, e.g. abcdisk.hd0 image
		initMedia();
		fl_init();
		if (fl_attach_media(media_read, media_write) != FAT_INIT_OK)
	    {
	    	printf("ERROR: Media attach failed\n");
			return -1;
	    }	
		break;

	case CMD32_buf:
		for (i=0;i<=255;i++) printf("%i ",abcMem[SDBUF+i]);
		break;

	case CMD32_rm:
		if (fl_remove("/file.bin") < 0)
		{ 
			printf("ERROR: Delete file failed\n"); 
			return -1;
		}
		break;

	case CMD32_secr:
		if(argc != 2) goto BAD_PARAM;
		if (sscanf(argv[1],"%lu",&sector)>0) 
		{
			printf("\r\nSector: %lu\r\n---------------------------------------------\r\n",sector);
			readSECTOR(sector,scBuf);
			for (i=0;i<512;i++)
				printf("%i ",scBuf[i]);
			printf("\r\n-----------------------------------------------------\r\n");
		}
		else 
		{ 
			printf("ERROR: Wrong sector number\n"); 
			return -1;
		}
		break;

	case CMD32_cr:
		file = fl_fopen("/cr.txt", "w+");
		if (file)
		{
			// Write some data
			fl_fwrite(d2,4,1,file);
			for (i=0;i<100;i++) scBuf[i]=0;
			{
				if (fl_fwrite(scBuf,100,1, file) != 100)
					printf("ERROR: Write file failed\n");
		    }
		}
		else
		{
			printf("ERROR: Create file failed\n");
		}
		// Close file
		printf("\r\n fflush %d",fl_fflush(file));
		printf("\r\nfseek: %i/fwrite %i ",fl_fseek(file,6,SEEK_SET),fl_fwrite(d2,4,1,file));
		printf("\r\n fflush %d",fl_fflush(file));
		printf("fseek: %i/fwrite %i ",fl_fseek(file,60,SEEK_SET),fl_fwrite(d2,4,1,file));
		printf("\r\n fflush %d",fl_fflush(file));
		printf("fseek: %i/fwrite %i ",fl_fseek(file,89,SEEK_SET),fl_fwrite(d2,4,1,file));
		printf("fseek: %i/read %i \r\n",fl_fseek(file,0,SEEK_SET),fl_fread(scBuf,100,1,file));
		printf("\r\n fflush %d",fl_fflush(file));
		for (i=0;i<100;i++) printf("%i ",scBuf[i]);
		{
			fl_fclose(file);
		}
		break;

	case CMD32_cat:
		if(argc != 2) goto BAD_PARAM;
		file = fl_fopen(argv[1], "r");
		if (file)
		{
			j=fl_fgetc(file);
			while (j!=EOF)	
			{
				printf("%c", (unsigned char) j);
				j=fl_fgetc(file);
			}  
			// Close file
			fl_fclose(file);
		}
		break;

	case CMD32_wifi:
		testWifi = True;
		printf("Wifi test ON, '>' to quit\r\n");
		break;

	case CMD32_peek:
		if((argc != 2) || (sscanf(argv[1],"%ld",&j)==0)) goto BAD_PARAM; //sscanf crash if argv[] not valid!
		j &= 65535; 
		printf("abcMem[%ld] is %u, WE=%u\r\n", j, abcMem[j], WEmap[j>>8]);
		break;

	case CMD32_poke:
		if((argc != 3) || (sscanf(argv[1],"%ld",&i)==0) || (sscanf(argv[2],"%ld",&j)==0)) goto BAD_PARAM; 
		i &= 65535;
		abcMem[i] = (uint8) j;
		printf("abcMem[%u] = %u\r\n", (uint16) i, (uint8) j);
		break;

	case CMD32_iopr:
		if(argc != 2) goto BAD_PARAM;
		j = *argv[1];
		if(j<'1')
		{
			ioprint = 0;
			printf("ioprint_OFF\n\r");
		}
		else
		{
			ioprint = j - '0';
			printf("ioprint_%d\n\r", ioprint);
		}
		break;

	case CMD32_iost:		//for testing io strobes without ABC80
		if(argc != 2) goto BAD_PARAM;
		j = *argv[1];
		/*
		if(c=='1')
		{	printf("io strobe test INT3\r\n");
			IFS0SET = 1<<15;
		}
		else if(c=='2')
		{	printf("io strobe test INT4\r\n");
			IFS0SET = 1<<19;
		}
		else if(c=='3')
		{	printf("io strobe test CN\r\n");
			IFS1SET = 1;
		}
		else if(c=='4')
		{	printf("MEMWR strobe test INT2\r\n");
			IFS0SET = 1<<11;
		}
		else
		{
			printf("Bad param");
		}
		*/
		printf("Test removed");
		break;

	case CMD32_ledr:
		if(argc != 2) goto BAD_PARAM;
		j = *argv[1];
		if(j == '1')
		{
			LEDRED_ON;
			printf("LEDRED_ON\r\n");
		}
		else
		{
			LEDRED_OFF;
			printf("LEDRED_OFF\r\n");
		}
		break;

	case CMD32_ledw:
		if(argc != 2) goto BAD_PARAM;
		j = *argv[1];
		if(j == '1')
		{
			LEDWHT_ON;
			printf("LEDWHT_ON\r\n");
		}
		else
		{
			LEDWHT_OFF;
			printf("LEDWHT_OFF\r\n");
		}
		break;

	case CMD32_ledg:
		if(argc != 2) goto BAD_PARAM;
		j = *argv[1];
		if(j=='1')
		{
			LEDGRN_ON;
			printf("LEDGRN_ON\r\n");
		}
		else
		{
			LEDGRN_OFF;
			printf("LEDGRN_OFF\r\n");
		}
		break;

	case CMD32_ledy:
		if(argc != 2) goto BAD_PARAM;
		j = *argv[1];
		if(j=='1')
		{
			LEDYEL_ON;
			printf("LEDYEL_ON\r\n");
		}
		else
		{
			LEDYEL_OFF;
			printf("LEDYEL_OFF\r\n");
		}
		break;

	case CMD32_boot:
		SoftReset();
		break;

	default:
		printf("Unknown command\r\n");
		return -1;
BAD_PARAM:
		printf("Bad param");
		return -1;
		break;

	} //switch

	return 0;
}

//end




