//Handle the UART transmit/receive buffers.
//UART2 can use hardware CTS/RTS.
//UART1 need software controlled CTS/RTS or DMA.

#include <plib.h>
#include "common.h"
#include "uarts.h"
#include "abcMemory.h"
#include "usb/usb_function_cdc.h"

//ABC SER is mapped to UART2 (Wifi)
//ABC COM can be mapped to UART1 (TTL) or USB
//Debug printf to PC can use USB or UART1
//Let's use USB for debug and hopefully USB keyboard in the future
#define ABC_USE_USB		0
#define DEBUG_USE_USB	1
//
#define ABC_USE_UART1	1
#define DEBUG_USE_UART1	0
//
#define UART1_USE_FTDI	0
//PIC32 inverted RX/TX levels matches FTDI non-inverted levels!
//If using a standard TTL-RS232 adapter, then set to 0.

//------------- USB uart --------------------------------
//interface between usb linear buffers and our circular buffers
#if(ABC_USE_USB)
	volatile uint8 *usbRxBuf = &abcMem[COMIN_BUF];
	volatile uint8 *usbTxBuf = &abcMem[COMOUT_BUF];
	#define usbRxHead	abcMem[COMIN_HEAD] 
	#define usbRxTail	abcMem[COMIN_TAIL]
	#define usbTxHead	abcMem[COMOUT_HEAD]
	#define usbTxTail	abcMem[COMOUT_TAIL]
#else
	uint8 usbRxBuf[256], usbTxBuf[256];
	uint8 usbRxHead, usbRxTail, usbTxHead, usbTxTail;
#endif

void pollUsbUart(void)
{	//This function is _only_ called from ProcessIO() in usb.c
	static char usbDataBuf[256];	//transfer buffer must be static
	uint8 i,j,n;

	if (! USBUSARTIsTxTrfReady()) return; //usbDataBuf not free to use

/*
//for testing usb uart 
	i = (uint8) getsUSBUSART(usbDataBuf, 255);
	if (i != 0)
		mUSBUSARTTxRam(usbDataBuf, i);
	return;
*/

	i = usbRxTail - usbRxHead - 1; //bytes free in Rx buffer
	if (i != 0)
	{	//read from usb to linear buffer and copy to circular buffer
		i = getsUSBUSART(usbDataBuf, i);
		n = usbRxHead;
		for(j=0; j<i; j++)
		{
			*(usbRxBuf + n++) = usbDataBuf[j];
		}
		usbRxHead = n;
		//return;
	}

	i = usbTxHead - usbTxTail;	//bytes to send in Tx buffer
	if (i != 0)
	{	//copy to linear buffer and send to usb
		n = usbTxTail;
		for(j=0; j<i; j++)
		{
			usbDataBuf[j] = *(usbTxBuf + n++);
		}
		usbTxTail = n;
		mUSBUSARTTxRam((unsigned char*) usbDataBuf, i); //usbDataBuf not free until all was sent
	}
}

void flushUsbRx(void)
{
	usbRxTail = usbRxHead;
}

void flushUsbTx(void)
{
	usbTxTail = usbTxHead;
}

uint8 bytesInUsbRx(void)
{
	return (usbRxHead - usbRxTail);
}

uint8 spaceInUsbTx(void)
{
	return (usbTxTail - usbTxHead - 1);
}

uint8 getCharUsb(void)
{
	if (usbRxHead != usbRxTail) return *(usbRxBuf + usbRxTail++);
	else return 0;
}

Bool putCharUsb(uint8 c)
{
	uint8 i = usbTxHead + 1;
	if(i == usbTxTail)
	{	//buffer full, try to call usb service
		usb();
	}
	if (i == usbTxTail) return 0;	//Maybe usb not connected.
	*(usbTxBuf + usbTxHead) = c;
	usbTxHead = i;
	return 1;
}


//----------------- UART2 -----------------------------------
//Intended for WiFly module
volatile uint8 *rxBuf2 = &abcMem[SERIN_BUF];
volatile uint8 *txBuf2 = &abcMem[SEROUT_BUF];
volatile uint8 *statusUart2 = &abcMem[SERERR];
#define rxHead2	abcMem[SERIN_HEAD]
#define rxTail2	abcMem[SERIN_TAIL]
#define txHead2	abcMem[SEROUT_HEAD]
#define txTail2	abcMem[SEROUT_TAIL]
uint8 rxHead2val;	//for debug
uint8 txHead2val;	//for debug
uint8 txTail2val;	//for debug

void serviceTx2(void)
{	//if anything to send and UART ready, then send
	while(txTail2 != txHead2)
	{
		if(U2STAbits.UTXBF) break;
		U2TXREG = *(txBuf2 + txTail2++);	
	}
	txTail2val = txTail2;
}

void pollUart2(void)
{	//UART2 use flow control in hardware

	//read incoming and put in rxBuffer, check for UART errors.
	while(U2STAbits.URXDA)
	{
		rxHead2val = rxHead2;
		if((uint8)(rxHead2val+1) == rxTail2) break;
		*(rxBuf2 + rxHead2val++) = U2RXREG;
		rxHead2 = rxHead2val;
		*statusUart2 |= (U2STA & 0x0f);
	}
	if(U2STAbits.OERR) U2STAbits.OERR=0;
	
	//if anything to send and UART ready, then send
	serviceTx2();
}

void flushRx2(void)
{
	rxTail2 = rxHead2;
}

void flushTx2(void)
{
	txTail2 = txHead2;
}

uint8 bytesInRx2(void)
{
	return (rxHead2 - rxTail2);
}

uint8 spaceInTx2(void)
{
	return (txTail2 - txHead2 - 1);
}

uint8 getCharRx2(void)
{
	if (rxHead2 != rxTail2) 
	{
		return *(rxBuf2 + rxTail2++);
	}
	else return 0;
}

Bool putCharTx2(uint8 c)
{	
	serviceTx2();
 	txHead2val = txHead2;
	*(txBuf2 + txHead2val++) = c;
	if (txHead2val == txTail2) return 0;	//overflow
	txHead2 = txHead2val;
	return 1;
}


//------------------ UART1 -----------------------
#if(ABC_USE_UART1)
	volatile uint8 *rxBuf1 = &abcMem[COMIN_BUF];
	volatile uint8 *txBuf1 = &abcMem[COMOUT_BUF];
	volatile uint8 *statusUart1 = &abcMem[COMERR];
	#define rxHead1	abcMem[COMIN_HEAD] 
	#define rxTail1	abcMem[COMIN_TAIL]
	#define txHead1	abcMem[COMOUT_HEAD]
	#define txTail1	abcMem[COMOUT_TAIL]
#else
	uint8 rxBuf1[256], txBuf1[256];
	uint8 rxHead1, rxTail1, txHead1, txTail1;
	uint8 status_uart1;
	uint8 *statusUart1 = &status_uart1;
#endif
#define RTS1	LATCbits.LATC13

#define RTS1_BSY	LATCSET = (1<<13)
#define RTS1_RDY	LATCCLR = (1<<13)
#define CTS1_BSY	(PORTCbits.RC14)
//Strange that FTDI inversion is only RX/TX (not RTS/CTS)

Bool serviceTx1(void)
{	//if anything to send and UART ready, then send
	if(txTail1 == txHead1) return 0;
	do
	{
		if(U1STAbits.UTXBF || CTS1_BSY) break;
		U1TXREG = *(txBuf1 + txTail1++);
	}while(txTail1 != txHead1);
	return 1;
}

void pollUart1(void)
{
	uint16 i;
	//Uart1 Rx handled by DMA. Check for errors and handle RTS.
	if(! DCH0CONbits.CHBUSY)
	{	//UART error occured, DMA gets disabled
		RTS1_BSY;
		if(rxTail1==rxHead1)
		{	//all good characters read, let's reinitiate
			DCH0CON=0x0010;		// channel off, priority 0, no chaining, auto-reenable
			while(U1STAbits.URXDA)
			{
				IFS0CLR = (1<<26) | (1<<27) | (1<<28);	// clear UART1 error flag
				i = U1RXREG;
				*statusUart1 |= U1STA;
			}
			U1STAbits.OERR=0;
			rxTail1 = rxHead1 = 0;	//DMA pointer will reset
			DCH0INT=0; // clear existing events, disable all dma0 interrupts
			DCH0CONbits.CHEN = 1;
		}
	}
	else
	{
		rxHead1 = DCH0DPTR;
		if(((uint8) (rxHead1 - rxTail1)) > 128) RTS1_BSY;
		else RTS1_RDY;
	}

	serviceTx1();
}

void flushRx1(void)
{
	rxTail1 = rxHead1;
}

void flushTx1(void)
{
	txTail1 = txHead1;
}

uint8 bytesInRx1(void)
{
	return (rxHead1 - rxTail1);
}

uint8 spaceInTx1(void)
{
	return (txHead1 - txTail1 - 1);
}

uint8 getCharRx1(void)
{
	if (rxHead1 != rxTail1) return *(rxBuf1 + rxTail1++);
	else return 0;
}

Bool putCharTx1(uint8 c)
{	uint8 i;
	serviceTx1();
 	i = txHead1;
	*(txBuf1 + i++) = c;
	if (i == txTail1) return 0;	//not done
	txHead1 = i;
	return 1;
}


//-------------- debug comm -----------------------
void flushPCRx(void)
{
	#if DEBUG_USE_USB
		flushUsbRx();
	#elif DEBUG_USE_UART1
		flushRx1();
	#endif
}

Bool DataRdyPC(void)
{	//check rxBuffer
	#if DEBUG_USE_USB
		return bytesInUsbRx();
	#elif DEBUG_USE_UART1
		return bytesInRx1();
	#else
		return 0;
	#endif
}

char getcPC(void)
{	//read from rxBuffer
	#if DEBUG_USE_USB
		return getCharUsb();
	#elif DEBUG_USE_UART1
		return getCharRx1();
	#else
		return 0;
	#endif
}

Bool putcPC(char c)
{	//put in txBuffer, return 0 if not done
	#if DEBUG_USE_USB
		return putCharUsb(c);
	#elif DEBUG_USE_UART1
		return putCharTx1(c);
	#else
		return 0;
	#endif
}

void _mon_putc(char c)
{	// for printf routine
	putcPC(c);
}

//-----------------------------------------------------------------------
void sendHexPC(uint8 c)
{
	uint8 i = (c>>4) + '0';
	if(i>'9') i += ('a' - '9' - 1);
	putcPC(i);
	i = (c&0x0f) + '0';
	if(i>'9') i += ('a' - '9' - 1);
	putcPC(i);
}


/******************************************************************************  
* Function:    unsigned int VirtToPhys(const void* p) * PreCondition:    None 
* Input:        p   - pointer to be converted.  
* Output:      a physical address corresponding to the virtual input pointer 
* Side Effects:    None  
Overview:       Virtual to physical translation helper. 
 ******************************************************************************/ 
unsigned int VirtToPhys(const void* p) 
//--------------------------------------------------------------------------- 
{ 	//could use macro KVA_TO_PA(v) in kmem.h instead /A.J
	return (int)p<0?((int)p&0x1fffffffL):(unsigned int)((unsigned char*)p+0x40000000L); 
}


//--------------------------------------------------------
void uart_init(void)
{
	#define FPB 			(80000000ul)
	#define UART1_BAUDRATE 	(115200ul)
	#define UART2_BAUDRATE 	(9600ul)

	//U2MODE = 0;	
	U2MODE=0x0200;	//RTS/CTS, n81.
	U2STA=0;	//IRQ when TX empty or RX 1 char
	U2BRG = (0.5 + FPB/16.0/UART2_BAUDRATE - 1);
	U2MODESET = 0x8000;		//UART ON
	U2STASET = 0x1400;		//RXEN, TXEN

	RTS1_BSY;
	U1MODE = 0;		//No hardware flow control, n81.
	U1STA = 0;		//IRQ when TX empty or RX 1 char
	U1BRG = (0.5 + FPB/16.0/UART1_BAUDRATE - 1);
	#if(UART1_USE_FTDI)
		U1MODESET = 0x0010;		//RX inverted.
		U1STASET = 0x2000;		//TX inverted.
	#endif
	U1MODESET = 0x8000;		//UART ON
	IFS0CLR = (1<<26) | (1<<27) | (1<<28);	// clear UART1 interrupts
	U1STASET = 0x1400;		//RXEN, TXEN

	//DMA setup, Family Reference Manual #31.
	DCH0CON=0;	DCH1CON=0;	DCH2CON=0;	DCH3CON=0;
	DCH4CON=0;	DCH5CON=0;	DCH6CON=0;	DCH7CON=0;
	IFS1CLR=0x00ff0000;	//clear all DMA IF
	DMACON=0x00008000;	// enable the DMA controller
	DCH0CON=0x0010;		// channel off, priority 0, no chaining, auto-reenable
	DCH0ECON=(26<<16) | (27<<8)| 0x18; // quit irq is UART Error, start irq is UART1 RX, pattern match disabled
	DCH0SSA = KVA_TO_PA(&U1RXREG); // transfer source physical address
	DCH0DSA = KVA_TO_PA(rxBuf1); // transfer destination physical address
	DCH0SSIZ = 1; // source size is 1 byte
	DCH0DSIZ=256; // destination size at most 256 bytes
	DCH0CSIZ=1; // one byte per UART transfer request
	DCH0INT=0; // clear existing events, disable all dma0 interrupts
	//IFS0SET = (1<<26) | (1<<27) | (1<<28);	// for testing dma restore
	DCH0CONbits.CHEN=1; // turn channel on
}

//end

